import ipywidgets as widgets # Loads the Widget framework.
from IPython.core.magics.namespace import NamespaceMagics # Used to query namespace.
import os
import shutil
import re
import psutil
import subprocess as sp
import os
from threading import Thread , Timer
import sched, time
import json

class Variable_data(object):
    instance = None

    def __init__(self, ipython):
        """Public constructor."""
        if Variable_data.instance is not None:
            raise Exception("""Only one instance of the Variable Inspector can exist at a 
                time.  Call close() on the active instance before creating a new instance.
                If you have lost the handle to the active instance, you can re-obtain it
                via `VariableInspectorWindow.instance`.""")
        
        Variable_data.instance = self
        self.closed = False
        self.namespace = NamespaceMagics()
        ipython.user_ns_hidden['widgets'] = widgets
        ipython.user_ns_hidden['NamespaceMagics'] = NamespaceMagics
        self.namespace.shell = ipython.kernel.shell

    def close(self):
        """Close and remove hooks."""
        if not self.closed:
            #self._ipython.events.unregister('post_run_cell', self.get_value)
            #self._box.close()
            self.closed = True
            Variable_data.instance = None

    def get_value(self):
        
        values = self.namespace.who_ls()
        return values
    
    def save_data(self,data,file_name):
        if not os.path.exists('experiment'):
            os.makedirs('experiment')
        data.to_csv(f"experiment/{file_name}.csv",index=False)
        

class Monitor(Thread):
    def __init__(self,delay,path):
        super(Monitor,self).__init__()
        self.stopped=False
        self.delay=delay
        self.path=path
        self.memory={'gpu':[],'ram':[]}
        self.base_ram=self.get_ram_usage()
        self.base_gpu=self.get_gpu_memory()
        self.start()
    
    def run(self):
        
        
        while not self.stopped:
            gpu=self.get_gpu_memory()
            ram=self.get_ram_usage()
            abs_gpu=0 if gpu-self.base_gpu<0 else gpu-self.base_gpu
            abs_ram=0 if ram-self.base_ram<0 else ram-self.base_ram
            self.memory['gpu'].append(abs_gpu)
            self.memory['ram'].append(abs_ram)
            time.sleep(self.delay)
            
    def get_ram_usage(self):
        
        """
        Obtains the absolute number of RAM bytes currently in use by the system.
        :returns: System RAM usage in bytes.
        :rtype: int

        """
        return int(int(psutil.virtual_memory().total - psutil.virtual_memory().available) / 1024 / 1024)


    def get_gpu_memory(self):
        
        output_to_list = lambda x: x.decode('ascii').split('\n')[:-1]
        ACCEPTABLE_AVAILABLE_MEMORY = 1024
        COMMAND = "nvidia-smi --query-gpu=memory.used --format=csv"
        try:
            memory_use_info = output_to_list(sp.check_output(COMMAND.split(),stderr=sp.STDOUT))[1:]
        except sp.CalledProcessError as e:
            raise RuntimeError("command '{}' return with error (code {}): {}".format(e.cmd, e.returncode, e.output))
        memory_use_values = [int(x.split()[0]) for i, x in enumerate(memory_use_info)]

        return memory_use_values[0]
    
    def stop(self):
        total_ram_consumption=max(self.memory['ram'])-min(self.memory['ram'])
        total_gpu_consumption=max(self.memory['gpu'])-min(self.memory['gpu'])
        self.memory['max_gpu_consumption']=total_gpu_consumption
        self.memory['max_ram_consumption']=total_ram_consumption
        if not os.path.exists(f"{self.path}/memory_info/"):
            os.makedirs(f"{self.path}/memory_info/")
        a_file = open(f"{self.path}/memory_info/memory_metrics.json", "w")
        a_file = json.dump(self.memory, a_file)
        self.stopped=True
        
def copy(src,dst):
    path="/".join(dst.split("/")[:-1])
    if re.search('[a-zA-Z]',path):
        if not os.path.exists(path):
            os.makedirs(path)
    shutil.copy(src,dst)

def save_dict(dict_,file_type,file_path):
    #path="/".join(file_path.split("/")[:-1])
    #file_name=file_path.split("/")[-1]
    file_type=file_type.lower()
    types=["performance","prediction"]
    if file_type in types:
        file_path_type=file_path+"/"+file_type
        if not os.path.exists(file_path_type):
            os.makedirs(file_path_type)
        a_file = open(file_path_type+f"/{file_type}.json", "w")
        a_file = json.dump(dict_, a_file)
    elif(file_type=="none"):
        a_file = open(file_path, "w")
        a_file = json.dump(dict_, a_file)
    else:
        print("file type not one of [performance,prediction,none]")


#!/usr/bin/env python
# coding: utf-8

# In[1]:


from flask import Flask,render_template,url_for,request,redirect
import plotly
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import os
import json
import pandas as pd
import numpy as np
from sklearn.metrics import roc_curve, roc_auc_score,precision_recall_curve,confusion_matrix
import jinja2
import plotly.figure_factory as ff
from pathlib import Path
import shutil
import pathlib
import glob
import re
#from ml_track_tool.flask_app import create_application


# In[2]:


pd.set_option('colheader_justify', 'center')


# In[3]:


def create_application(path):
    app=Flask(__name__)
    current_path=pathlib.Path(__file__).parent.resolve()
    my_loader = jinja2.ChoiceLoader([
        app.jinja_loader,
        jinja2.FileSystemLoader(['/flaskapp/userdata',
                                 f"{current_path}/templates"]),
    ])
    app.jinja_loader = my_loader
    @app.route('/',methods=["POST","GET"])
    def home():   
        if request.method=='POST':
            experiment_folder=request.form['experiment']
            if os.path.exists(f"{path}/{experiment_folder}/notes/table_{experiment_folder}.html"):
                shutil.copy(f"{path}/{experiment_folder}/notes/table_{experiment_folder}.html",f"{current_path}/templates/")
            return redirect(url_for("home2",experiment=experiment_folder))
        
        else:
            folders=os.listdir(path)
            return render_template("experiment_page.html",folders=folders)
        
    @app.route('/home2/<experiment>',methods=["POST","GET"])
    def home2(experiment):
        if request.method=='POST':
        
            notes=request.form['text']
            text_file = open(f"{path}/team_notes/notes.txt", "w")
            #write string to file
            text_file.write(notes)
            #close file
            text_file.close()
            download_path=f"{Path.home()}/Downloads/*.csv".replace("\\","/")
            latest_file=get_latest_file(download_path)
            if latest_file:
                df=pd.read_csv(latest_file).iloc[:,1:]
                df=df.replace(np.nan,"")
                html=df.to_html()
                html=html.replace('<table border="1" class="dataframe">','<table border="1" width="100%" contenteditable=true  class="dataframe" id="table_">')
                with open(f"{current_path}/templates/table_{experiment}.html", 'w') as f:
                    f.write(html)
                os.remove(latest_file)
            return redirect(url_for("home2",experiment=experiment))
        else:
            exp_folder=experiment
            table_path=f"table_{exp_folder}.html"
            
            if not os.path.exists(f"{current_path}/templates/{table_path}"):
                table_path="none"
            
            memory_path=f"{path+'/'+exp_folder}/memory_info/memory_metrics.json"
            history_path=f"{path+'/'+exp_folder}/performance/history.json"
            prediction_path=f"{path+'/'+exp_folder}/prediction/prediction.json"

            memory_file_path_exists=False
            history_file_path_exists=False
            prediction_file_path_exists=False
            memory_dict={}
            history_dict={}
            pred_dict={}
            if os.path.exists(memory_path):
                memory_file = open(memory_path, "r")
                memory_dict=json.load(memory_file)
                memory_file_path_exists=True
            if os.path.exists(history_path):
                history_file = open(history_path, "r")
                history_dict=json.load(history_file)
                history_file_path_exists=True
            if os.path.exists(prediction_path):
                
                pred_file = open(prediction_path, "r")
                pred_dict=json.load(pred_file)
                prediction_file_path_exists=True
            if not os.path.exists(f"{path}/team_notes/notes.txt"):
                os.mkdir(f"{path}/team_notes/notes.txt")
            notes_path=f"{path}/team_notes/notes.txt"
            with open(notes_path) as f:
                contents = f.read()
                
            fig=plots(memory_dict,history_dict,pred_dict,memory_file_path_exists,history_file_path_exists,prediction_file_path_exists)
            graphJSON = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return render_template('visualization.html', graphJSON=graphJSON,user_notes=contents,table_path=table_path)
    def plot_helper(fig,text,row,col):
            fig.add_trace(go.Scatter(
            x=[0],
            y=[0],
            mode="lines+markers+text",showlegend=False,
            text=[text],
            textposition="bottom center",
            textfont=dict(
                family="sans serif",
                size=20,
                color="crimson"
            ),
        ),row=row[0],col=col[0])
            fig.add_trace(go.Scatter(
            x=[0],
            y=[0],
            mode="lines+markers+text",showlegend=False,
            text=[text],
            textposition="bottom center",
            textfont=dict(
                family="sans serif",
                size=20,
                color="crimson"
            ),
        ),row=row[1], col=col[1])
    def plot_confusion_matrix(fig,y_true,y_pred,x,y,row,col,labels):
        confusion_matrix_=confusion_matrix(y_true,y_pred)
        trace1 = ff.create_annotated_heatmap(z = confusion_matrix_,
                                 x = labels,
                                 y = labels,
                                 showscale  = False,name = "matrix")
        fig.add_trace(go.Heatmap(trace1.data[0]),row=row,col=col)
        
    def plots(memory_dict,history_dict,pred_dict,memory_file_path_exists,history_file_path_exists,prediction_file_path_exists):
             
        fig = make_subplots(rows=4, cols=2,subplot_titles=("RAM Consumption","GPU Consumption","Train & Validation loss","Train & Validation accuracy","ROC Curve","Precision-Recall Curve","Confusion matrix"))
        if(memory_file_path_exists):
            fig.add_trace(
            go.Scatter(y=memory_dict['ram'],name="RAM"),
            row=1, col=1)

            fig.add_trace(
                go.Scatter(y=memory_dict['gpu'],name="GPU"),
                row=1, col=2)
        else:
            plot_helper(fig,"Memory info unavailable",[1,1],[1,2])
            
        if(history_file_path_exists):
            keys=list(history_dict.keys())
            x3=[i for i in range(len(history_dict[keys[0]]))]
            fig.add_trace(
                go.Scatter(x=x3,y=history_dict[keys[0]],name=keys[0]),
                row=2, col=1)
            fig.add_trace(
                go.Scatter(x=x3,y=history_dict[keys[2]],name=keys[2]),
                row=2, col=1)
            fig.add_trace(
                go.Scatter(x=x3,y=history_dict[keys[1]],name=keys[1]),
                row=2, col=2)
            fig.add_trace(
                go.Scatter(x=x3,y=history_dict[keys[3]],name=keys[3]),
                row=2, col=2)
        else:
            plot_helper(fig,"Performance metrics info unavailable",[2,2],[1,2])        
        
        if prediction_file_path_exists:
            
            y_true_val=np.array(pred_dict['y_true'])
            y_pred_proba=np.array(pred_dict['y_pred'])
            
            if np.ndim(y_pred_proba)>1:
                y_pred=np.argmax(y_pred_proba,axis=1)
                labels_id=[i for i in range(y_pred_proba.shape[1])]
                for i in range(y_true_val.shape[1]):
                    y_true = y_true_val[:, i]
                    y_score = y_pred_proba[:, i]

                    fpr, tpr, _ = roc_curve(y_true, y_score)
                    auc_score = roc_auc_score(y_true, y_score)
                    precision, recall, thresholds = precision_recall_curve(y_true, y_score)
                    name = f"{i} (AUC={auc_score:.2f})"
                    fig.add_trace(go.Scatter(x=fpr, y=tpr, name=name, mode='lines'),row=3,col=1)
                    fig.add_trace(go.Scatter(x=[0,1], y=[0,1], line = dict(color='royalblue', dash='dash'),showlegend=False),row=3,col=1)
                    fig.add_trace(go.Scatter(x=recall, y=precision, fill='tozeroy',name=f"Precision-Recall Curve:{i}"),row=3,col=2)
                    fig.add_trace(go.Scatter(x=[1,0], y=[0,1], line = dict(color='royalblue', dash='dash'),showlegend=False),row=3,col=2)
                
                plot_confusion_matrix(fig,y_true,y_pred,[0,1],[0,1],4,1,labels_id)
            else:
                y_pred=np.round(y_pred_proba)
                labels_id=[0,1]
                fpr, tpr, _ = roc_curve(y_true, y_pred_proba)
                auc_score = roc_auc_score(y_true, y_pred_proba)
                precision, recall, thresholds = precision_recall_curve(y_true, y_score)
                fig.add_trace(go.Scatter(x=fpr, y=tpr, fill='tozeroy'),row=3,col=2) # fill down to xaxis
                fig.add_trace(go.Scatter(x=[0,1], y=[0,1], line = dict(color='royalblue', dash='dash'),showlegend=False),row=3,col=2)
                fig.add_trace(go.Scatter(x=recall, y=precision, fill='tozeroy',name=f"Precision-Recall Curve"),row=3,col=2) # fill down to xaxis
                fig.add_trace(go.Scatter(x=[1,0], y=[0,1], line = dict(color='royalblue', dash='dash'),showlegend=False),row=3,col=2)
                plot_confusion_matrix(fig,y_true,y_pred,[0,1],[0,1],4,1,labels_id)
        
        else:
            plot_helper(fig,"Prediction info unavailable",[3,3],[1,2])
           
        
        fig['layout']['xaxis']['title']='Seconds'
        fig['layout']['yaxis']['title']='RAM Consumption (MB)'
        fig['layout']['xaxis2']['title']='Seconds'
        fig['layout']['yaxis2']['title']='GPU Consumption (MB)'
        fig['layout']['xaxis3']['title']='epochs'
        fig['layout']['yaxis3']['title']='Loss'
        fig['layout']['xaxis4']['title']='epochs'
        fig['layout']['yaxis4']['title']='Accuracy'
        fig['layout']['xaxis5']['title']='False Positive Rate'
        fig['layout']['yaxis5']['title']='True Positive Rate'
        fig['layout']['xaxis6']['title']='Recall'
        fig['layout']['yaxis6']['title']='Precision'
        fig.update_layout(autosize=False,width=1300,height=1500)
        return fig
    
    def get_latest_file(download_path):
        pattern=r'export_table_\(?'
        list_of_files = glob.glob(download_path)
        csv_files=list(map(lambda x:x.split("\\")[1],list_of_files))
        matched_files=list(filter(re.compile(pattern).match, csv_files))
        list_of_files=list(map(lambda x:f"{Path.home()}/Downloads/"+x,list(matched_files)))
        latest_file = max(list_of_files, key=os.path.getctime)
        return latest_file


    app.run()







import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix,precision_score,recall_score,f1_score,accuracy_score
from matplotlib.pyplot import figure
import scikitplot as skplt
import tensorflow as tf



def plot_metric(y_true,y_pred_proba):
    ''' y_true shape:(n_samples)
    
        y_pred shape: (n_samples,n_classes)
    '''
    y_pred=tf.argmax(y_pred_proba,axis=1)
    precision=precision_score(y_true,y_pred,average='binary')
    recall=recall_score(y_true,y_pred,average='binary')
    f1score=f1_score(y_true,y_pred)
    accuracy=accuracy_score(y_true,y_pred)

    metrics=f'''
            accuracy:{round(accuracy,2)}% 

            F1_score:{round(f1score,2)}%

            Precision:{round(precision,2)}%

            Recall: {round(recall,2)}%'''

    fig, ax = plt.subplots(1, 2, sharex='col', sharey='row',figsize=(16,7))

    ax[0].text(0.15,0.3,metrics,clip_on=True,fontdict={"size":20})
    ax[0].axis('off')


    skplt.metrics.plot_roc(y_true, y_pred_proba,ax=ax[1])
    plt.show()


import lime
from lime import lime_text
from lime.lime_text import LimeTextExplainer
import ipysheet

def lime_explainer(class_names,text,predict_proba):
    explainer = LimeTextExplainer(class_names = class_names) 
    if type(text)==list:
        for i in range(len(text)):
            exp = explainer.explain_instance(text[i], predict_proba)
            exp.show_in_notebook(text=text[i])
            
    elif type(text)==str:
        exp = explainer.explain_instance(text, predict_proba)
        exp.show_in_notebook(text=text)
        
    elif type(text).__name__=='Sheet':
        df=ipysheet.to_dataframe(text)
        cols=df.columns.tolist()
        text=df[df[cols[-1]]=="y"][cols[0]].values.tolist()
        for i in range(len(text)):
            exp = explainer.explain_instance(text[i], predict_proba)
            exp.show_in_notebook(text=text[i])



import pandas as pd
import ipywidgets as widgets # Loads the Widget framework.
from ipysheet import from_dataframe
import ipysheet

def filter_(sheet,query):
    
    df=ipysheet.to_dataframe(sheet)
    condition=query.split("&")
    value=list(map(lambda x:x.split("==")[1].strip(),condition))
    col=list(map(lambda x:x.split("==")[0].strip(),condition))[0]
    
    value=[int(val) if val.isdigit() else val for val in value ]

    df=df[df[col].isin(value)]
   
    sheet=from_dataframe(df)
    sheet.column_width=[10]*len(df.columns)
    return sheet

def create_column(n_cols,sheet):
    df=ipysheet.to_dataframe(sheet)
    for i in range(n_cols):
        df[f"col{i}"]=""
    sheet=from_dataframe(df)
    sheet.column_width=[10]*len(df.columns)
    return sheet